######################################################################
#
# IMAPMessage, manage IMAP messages for Webmail product
#
# This software is governed by a license. See 
# LICENSE.txt for the terms of this license.
#
######################################################################

from IMAPPasserelle import *
from IMAPMessage import *
import cStringIO
import mimetools
import MimeWriter

try:
    import mimify
    MIMIFY_LIB = 1
except:
    MIMIFY_LIB = 0

try:
    from email import Encoders
    from email.MIMEAudio import MIMEAudio
    from email.MIMEBase import MIMEBase
    from email.MIMEMultipart import MIMEMultipart
    from email.MIMEImage import MIMEImage
    from email.MIMEText import MIMEText
    from MessageTools import encoded_part, encoded_addr, parse_Addresses
    EMAIL_LIB = 1
except:
    EMAIL_LIB = 0


import re
import string
from DocumentTemplate.DT_Util import html_quote
from DocumentTemplate.DT_Var import newline_to_br
import quopri

DEFAULT_CHARSET = 'iso-8859-1'

class IMAPMessage:
    """ IMAP message """
    
    def __init__(self, IMAPId = "", flags = {}, expediteur = [],
                 size = 0, date = "", subject = "[pas de sujet]",
                 body = "", destinataires = {'to':[], 'cc':[], 'bcc':[]},
                 replyTo = '', attachements = [], headers = {},content_type = 'text/plain'
                 ):
        """ constructor"""

        self.headers = headers
        self.IMAPId = IMAPId
        self.flags = flags
        self.expediteur = expediteur
        self.size = size
        self.date = date
        self.subject = subject
        self.body = body
        self.destinataires = destinataires
        self.replyTo = replyTo
        self.attachements = attachements
        self.content_type = content_type
        
    def getIMAPId(self):
        """ return IMAP Id fo the message"""
        return self.IMAPId
    
    def getContentType(self):
        """ return content type of self """
        return self.content_type 
    
    def isRead(self):
        """ return the value of flag Seen (= Read) """
        return self.flags['read']

    def isAnswered(self):
        """ return the value of the flag Answered """
        return self.flags['answered']

    def isDeleted(self):
        """ return the value of the flag Deleted """
        return self.flags['deleted']

    def isFlagged(self):
        """ return the value of the flag Flagged """
        return self.flags['flagged']

    def getSize(self):
        """ return the size of the message """
        return self.size

    def getDate(self):
        """ return readable date of the message """
        return self.date

    def getSubject(self):
        """ return the subject """
        return self.subject

    def setSubject(self, subject):
        """ set the subject """
        self.subject = subject

    def getExpediteurName(self):
        """ return the sender name """
        return string.replace(self.expediteur[0], "\"", '')

    def getExpediteurMail(self):
        """ return the sender mail """
        return string.replace(self.expediteur[1], "\"", '')

    def getTo(self):
        """ return the 'to' field (list)"""
        return self.destinataires['to']

    def setTo(self, to):
        """ set to value """
        self.destinataires['to'] = to

    def getCC(self):
        """ return the 'cc' field (list)"""
        return self.destinataires['cc']

    def setCC(self, cc):
        """ set cc value """
        self.destinataires['cc'] = cc

    def getBCC(self):
        """ return the 'bcc' field (list)"""
        return self.destinataires['bcc']

    def setBCC(self, bcc):
        """ set bcc value """
        self.destinataires['bcc'] = bcc

    def getReplyTo(self):
        """ return the reply-to """
        return self.replyTo

    def setReplyTo(self,replyTo=''):
        """ set the reply-to """
        self.replyTo = replyTo

    def getBody(self):
        """ return the body of the message """
        return self.body

    def setBody(self, body):
        """ set the message body """
        self.body = body

    def setDestinataires(self, destinataires):
        """ set destinataires values"""
        self.destinataires = destinataires

    def getDestinataires(self):
        """ Return destinataire """
        return self.destinataires
    
    def getAttachements(self):
        """ return the attachement list"""
        return self.attachements

    def delAttachement(self, id):
        """ delete a attachement """
        #to do

    def edit(self, body, subject, date, expediteur, attachement, to, cc, bcc, copy_to_folder, save_draft):
        """ edit a message """        
        self.setBody(body)
        self.subject = subject
        self.date = date
        self.expediteur = expediteur
        self.attachements = attachement
        self.destinataires = {'to': to,'cc': cc,'bcc': bcc}

    def raw_message(self):
        """ Style rendering selector """
        if EMAIL_LIB:
            return self.email_raw_message()
        else:
            return self.RFC_raw_message()

    def email_raw_message(self):
        """
        Build a Email with email Library
        """
        # Create the enclosing (outer) message
        if self.getAttachements():
            outer = MIMEMultipart()
            body = MIMEText(self.getBody(), _charset = DEFAULT_CHARSET)
            outer.attach(body)
        else:
            outer = MIMEText(self.getBody(), _charset = DEFAULT_CHARSET)

        # Build header
        for header_key in self.headers.keys():
            if header_key == 'copy-to-folder' or header_key.lower() == 'subject':
                continue
            outer.add_header(string.capitalize(header_key), self.headers[header_key])
            
        sender = encoded_addr((encoded_part(self.expediteur[0]), self.expediteur[1]))

        To, Cc, ReplyTo = '', '', ''

        if self.getTo():
            To = parse_Addresses(self.getTo()[0])
        if self.getCC():
            Cc =  parse_Addresses(self.getCC()[0])
        if self.getReplyTo():
            ReplyTo = self.getReplyTo()
        subject = self.headers.get('subject', '') or self.headers.get('Subject', '')
        subject = encoded_part(subject)
        
        outer.add_header('From', sender)
        outer.add_header('To', To)
        if Cc:
            outer.add_header('Cc', Cc)
        if ReplyTo:
            outer.add_header('Reply-To', ReplyTo)
        outer.add_header('Subject', subject)
        outer.add_header('X-Mailer', 'WebMailv4')
            
        for attachement in self.getAttachements():
            filename = attachement.getFilename()
            # Guess the content type based on the file's extension.  Encoding
            # will be ignored, although we should check for simple things like
            # gzip'd or compressed files.
            ctype = attachement.getContentType()
            maintype, subtype = ctype.split('/', 1)
            if maintype == 'text':
                # Note: we should handle calculating the charset
                msg = MIMEText(attachement.getData(), _subtype=subtype, _charset = DEFAULT_CHARSET)
            elif maintype == 'image':
                msg = MIMEImage(attachement.getData(), _subtype=subtype)
            elif maintype == 'audio':
                msg = MIMEAudio(attachement.getData(), _subtype=subtype)
            else:
                msg = MIMEBase(maintype, subtype)
                msg.set_payload(attachement.getData())
                # Encode the payload using Base64
                Encoders.encode_base64(msg)
            # Set the filename parameter
            msg.add_header('Content-Disposition', 'attachment', filename=filename)
            outer.attach(msg)
        return outer.as_string()


    def RFC_raw_message(self):
        """
        Renders an RFC822 compliant message.
        (Old Rendering : use email lib instread)
        """

        rendered_message = cStringIO.StringIO()
        #writer = extended_mime_writer.extended_mime_writer(rendered_message)
        writer = MimeWriter.MimeWriter(rendered_message)
        sender = self.expediteur[0] + ' <' + self.expediteur[1] + '>'
        writer.addheader('From', sender)
        if self.getReplyTo() != '':
            writer.addheader('Reply-To', self.getReplyTo())
        for header_key in self.headers.keys():
            if header_key == 'copy-to-folder':
                continue
            if header_key == 'subject' and self.headers['subject']:
                suj = self.headers['subject']
                from cStringIO import StringIO
                infp = StringIO(suj)
                outfp = StringIO()
                quopri.encode(infp, outfp, 1)
                sujet = outfp.getvalue()
                sujet = string.replace(sujet, "\n", "")
                sujet = string.replace(sujet, " ", "_")
                
                if string.find(sujet, "?") == -1:
                    sujet = "=?iso-8859-1?Q?" + sujet + "?="
                else:
                    sujet = string.replace(sujet, "?", "=3F")
                    sujet = "=?iso-8859-1?Q?" + sujet + "?="
                    writer.addheader(header_key, sujet)
            else:
                writer.addheader(header_key, self.headers[header_key])

        for address_field in ['to', 'cc']:
            if self.destinataires[address_field]:
                for address in self.destinataires[address_field]:
                    writer.addheader(string.capitalize(address_field),
                                     mimify.mime_encode_header(address))
                    
        if self.getAttachements():
            writer.addheader('Mime-Version', '1.0')

        writer.addheader('X-Mailer', 'WebMailv4')

        writer.flushheaders()

        if self.getAttachements():
            writer.startmultipartbody('mixed')
            mime_part = writer.nextpart()
            mime_part.addheader('Content-Transfer-Encoding', 'quoted-printable')
            mime_part_writer = mime_part.startbody(self.getContentType(), [('charset', 'iso-8859-1')])
            #mimetools.encode(cStringIO.StringIO(self.body),mime_part_writer, 'quoted-printable')
            body = cStringIO.StringIO(self.body)
            encbody = cStringIO.StringIO()
            quopri.encode(body, encbody, 0)
            encbody.seek(0)
            mimetools.copyliteral(encbody, mime_part_writer)
            
            for attachment in self.getAttachements():
                mime_part = writer.nextpart()
                mime_part.addheader('Content-Transfer-Encoding','base64')

                # ?FIXME?
                # Let the attachment encoding decide?
                
                #tmp_headers = {'content-transfer-encoding':
                ##'base64'}
                ##tmp_headers.update(attachment.headers)
                mime_part_writer = mime_part.startbody(
                    attachment.content_type,
                    [('filename', attachment.filename),('name', attachment.filename)],
                    1)
                
                
                mime_part_writer.write(attachment.encode())
                
            else:
                writer.lastpart()

        else:
            writer._fp.write('Content-Transfer-Encoding: quoted-printable\n')
            body_writer = writer.startbody(self.getContentType(), [],{'Content-Transfer-Encoding':'quoted-printable'})
            self.body = '\n' + self.body
            body = cStringIO.StringIO(self.body)
            encbody = cStringIO.StringIO()
            quopri.encode(body, encbody, 0)
            encbody.seek(0)
            mimetools.copyliteral(encbody, body_writer)
            #mimetools.encode(body, body_writer, 'quoted-printable')
            
        return rendered_message.getvalue()


    def setRead(self, value):
        """ set read (Seen) flag"""
        

    def setAnswered(self, folderName):
        """ set Answered flag """

        
    def setDeleted(self, value):
        """ set deleted flag """

    def existAttachement(self):
        """ return true if message have one or more attachment """
        return len(self.attachements)


    def render_body(self, body, version='text', reply = 0):
        """ replace hexadecimal character with ascii"""
        
        if not body:
            body = ''

        if not EMAIL_LIB and MIMIFY_LIB: 
            body = mimify.mime_decode(body)
            body = string.replace(body, "=\r\n", "")
            # The attachmement process !
            body = re.sub(r'(==.*)(REL|ALT)(.*)(\n|\r)(Content-.*\n|\r){1,3}([\S\s]*)(--=.*)(REL|ALT)(?!--)([\S\s]*)', r'\6', body)

        mess = body
        
        if version == "enriched":
            mess = re.sub(r'<param>([\S]*)</param>', r' ', mess)
            for balise in ['<color>', '</color>', '<bold>', '</bold>', '<bigger>', '</bigger>', '<underline>', '</underline>', '<italic>', '</italic>', '<smaller>', '</smaller>', '<center>', '</center>', '<paraindent>', '</paraindent>', '<excerpt>', '</excerpt>','<fontfamily>', '</fontfamily>']:
                mess = string.replace(mess, balise, '')

        if version in ['text', 'enriched']:
            decoder = re.compile(r'(<html>)', re.IGNORECASE | re.DOTALL)
            _text = decoder.sub('<html>', mess)
            decoder = re.compile(r'(</html>)', re.IGNORECASE | re.DOTALL)
            _text = decoder.sub('<html>', _text)
            
            splittedText = _text.split('<html>')
            i = 0
            l = []
            
            for item in splittedText:
                if i % 2 == 0:
                    # Here a text
                    if reply:
                        t = '\n> ' + string.join(string.split(item, '\n'), '\n> ') + '\n\n'
                        l.append(t)
                    else:    
                        l.append(newline_to_br(self._URLsToLinksOfText(html_quote(item))))
                else:
                    # Here HTML
                    l.append(item)
                i += 1
            mess = ''
            i = 0
            for text in l:
                if i % 2 != 0:
                    a, b = '<html>', '</html>'
                else:
                    a, b = '' ,''
                mess = mess + a + text + b 
                i += 1
                
        return mess

    def _URLsToLinksOfText(self, mess):
        """ Change text URL in HTML link URL and change space to &nbsp; """
        if EMAIL_LIB:
            mess = string.replace(mess, ' ', '\0')
            p = re.compile('([^\0])\0', re.IGNORECASE | re.MULTILINE)
            r = p.sub(r'\1 ', mess)
            mess = string.replace(r, '\0', '&nbsp;')
            
        mess = re.sub(r'(http://\S*)(&quot;)', r'\1 \2 ', mess)
        # Change 'http or https link in html for display
        mess = re.sub(r'(http(s?)://\S*)', r'<a href="\1" target="_blank">\1</a>', mess)
        # Change 'mailto' in html display
        mess = re.sub(r'(mailto:(\S*)\?subject=(\S*))', r'<a href="edit_message?to=\2&subject=\3">\2</a>', mess)
        mess = re.sub(r'(mailto:(\S*))', r'<a href="edit_message?to=\2">\2</a>', mess)
        # Remove space of the begining
        mess = string.replace(mess, '/a> &quot; &gt;', '/a>&quot; &gt;')

        

        try:
            if not EMAIL_LIB:
                mess = self.processSpace(mess)
        except:
            pass
        return mess

    def get_parsed_body(self):
        """ return parsed body """
        try:

            if EMAIL_LIB:
                return  self.render_body(self.getBody(), 'text')
            
            if string.find(string.lower(self.headers['content-type']), 'text/html')!=(-1):
		return self.render_body(self.getBody(), 'html')
            elif string.find(self.headers['content-type'], 'text/enriched')!=(-1):
                return newline_to_br(self.render_body(self.getBody(), 'enriched'))
            else:
                return newline_to_br(self.render_body(self.getBody(), 'text'))
        except:
            return newline_to_br(self.render_body(self.getBody(), 'text'))
            
    def processSpace(self,mess):
        """ process space at the begining of the line and in the line """
        body = mess
        listLines = string.split(body, '\n')
        newListLines = []
        for line in listLines:
            currentLine = ""
            i = 0
            while len(line) > i and line[i] == " ":
                currentLine += "&nbsp;"
                i += 1
            newListLines.append(currentLine + line[i:])
        body = string.join(newListLines, '\n')
        body = string.replace(body, '\t', '&nbsp;&nbsp;&nbsp;&nbsp;')
        return body

    def decode_body(self):
        """ decode quopri """
        if EMAIL_LIB:
            return self.getBody()
        mess = mimify.mime_decode(self.getBody())
        mess = string.replace(mess, "=\r\n", "")
        return mess
